# Copyright 2014-2017 Canonical Limited.
#
# This file is part of charms.reactive
#
# charms.reactive is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charms.reactive is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.

import os
import traceback

from .flags import *  # noqa
from .relations import *  # noqa
from .endpoints import *  # noqa
from .decorators import *  # noqa
from .helpers import *  # noqa
from .patterns.request_response import *  # noqa

from . import bus
from . import flags
from . import helpers
from charmhelpers.core import hookenv
from charmhelpers.core import unitdata


# DEPRECATED: transitional imports for backwards compatibility
bus.StateList = flags.StateList
bus.State = flags.State
bus.set_state = flags.set_flag
bus.remove_state = flags.clear_flag
bus.get_state = flags.get_state
bus.get_states = flags.get_states
helpers.is_state = flags.is_flag_set
helpers.toggle_state = flags.toggle_flag
helpers.all_states = flags.all_flags_set
helpers.any_states = flags.any_flags_set


def main(relation_name=None):
    """
    This is the main entry point for the reactive framework.  It calls
    :func:`~bus.discover` to find and load all reactive handlers (e.g.,
    :func:`@when <decorators.when>` decorated blocks), and then
    :func:`~bus.dispatch` to trigger handlers until the queue settles out.
    Finally, :meth:`unitdata.kv().flush <charmhelpers.core.unitdata.Storage.flush>`
    is called to persist the flags and other data.

    :param str relation_name: Optional name of the relation which is being handled.
    """
    hook_name = hookenv.hook_name()
    restricted_mode = hook_name in ['meter-status-changed', 'collect-metrics']

    hookenv.log('Reactive main running for hook %s' % hookenv.hook_name(), level=hookenv.INFO)
    if restricted_mode:
        hookenv.log('Restricted mode.', level=hookenv.INFO)

    # work-around for https://bugs.launchpad.net/juju-core/+bug/1503039
    # ensure that external handlers can tell what hook they're running in
    if 'JUJU_HOOK_NAME' not in os.environ:
        os.environ['JUJU_HOOK_NAME'] = hook_name

    try:
        bus.discover()
        if not restricted_mode:  # limit what gets run in restricted mode
            hookenv._run_atstart()
        bus.dispatch(restricted=restricted_mode)
    except Exception:
        tb = traceback.format_exc()
        hookenv.log('Hook error:\n{}'.format(tb), level=hookenv.ERROR)
        raise
    except SystemExit as x:
        if x.code not in (None, 0):
            raise

    if not restricted_mode:  # limit what gets run in restricted mode
        hookenv._run_atexit()
    unitdata._KV.flush()



# Copyright 2014-2015 Canonical Limited.
#
# This file is part of charm-helpers.
#
# charm-helpers is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charm-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.

import unittest

import mock

from charmhelpers.core import unitdata
from charms import reactive


class TestReactiveMain(unittest.TestCase):
    @mock.patch.object(unitdata, '_KV')
    @mock.patch.object(reactive.bus, 'dispatch')
    @mock.patch.object(reactive.bus, 'discover')
    @mock.patch.object(reactive.hookenv, '_run_atstart')
    @mock.patch.object(reactive.hookenv, 'log')
    @mock.patch.object(reactive.hookenv, 'hook_name')
    @mock.patch.object(reactive.relations, 'relation_factory')
    def test_main(self, rel_factory, hook_name, log, _run_atstart, discover, dispatch, _KV):
        hook_name.return_value = 'hook_name'
        reactive.main()
        _run_atstart.assert_called_once_with()
        log.assert_called_once_with('Reactive main running for hook hook_name', level=reactive.hookenv.INFO)
        discover.assert_called_once_with()
        dispatch.assert_called_once_with(restricted=False)
        _KV.flush.assert_called_once_with()

        _KV.flush.reset_mock()
        discover.side_effect = SystemExit(1)
        self.assertRaises(SystemExit, reactive.main)
        assert not _KV.flush.called


class TestReactiveRestrictedMain(unittest.TestCase):
    @mock.patch.object(unitdata, '_KV')
    @mock.patch.object(reactive.bus, 'dispatch')
    @mock.patch.object(reactive.bus, 'discover')
    @mock.patch.object(reactive.hookenv, '_run_atstart')
    @mock.patch.object(reactive.hookenv, 'log')
    @mock.patch.object(reactive.hookenv, 'hook_name')
    def test_main(self, hook_name, log, _run_atstart, discover, dispatch, _KV):
        hook_name.return_value = 'meter-status-changed'
        reactive.main()
        _run_atstart.assert_not_called()
        log.any_call('Reactive restricted main running for hook meter-status-changed', level=reactive.hookenv.INFO)
        log.any_call('Restricted mode.', level=reactive.hookenv.INFO)
        discover.assert_called_once_with()
        dispatch.assert_called_once_with(restricted=True)
        _KV.flush.assert_called_once_with()

        _KV.flush.reset_mock()
        discover.side_effect = SystemExit(1)
        self.assertRaises(SystemExit, reactive.main)
        assert not _KV.flush.called



# Copyright 2014-2017 Canonical Limited.
#
# This file is part of charm-helpers.
#
# charm-helpers is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charm-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.

import importlib
import os
import sys
import errno
import subprocess
from itertools import chain
from functools import partial

from charmhelpers.core import hookenv
from charmhelpers.core import unitdata
from charms.reactive.trace import tracer


_log_opts = os.environ.get('REACTIVE_LOG_OPTS', '').split(',')
LOG_OPTS = {
    'register': 'register' in _log_opts,
}


class BrokenHandlerException(Exception):
    def __init__(self, path):
        message = ("File at '{}' is marked as executable but "
                   "execution failed. Only handler files may be marked "
                   "as executable.".format(path))
        super(BrokenHandlerException, self).__init__(message)


def _action_id(action, suffix=None):
    if hasattr(action, '_action_id'):
        return action._action_id
    parts = [
        action.__code__.co_filename,
        action.__code__.co_firstlineno,
        action.__code__.co_name,
    ]
    if suffix is not None:
        parts.append(suffix)
    return ':'.join(map(str, parts))


def _short_action_id(action, suffix=None):
    if hasattr(action, '_short_action_id'):
        return action._short_action_id
    filepath = os.path.relpath(action.__code__.co_filename, hookenv.charm_dir())
    parts = [
        filepath,
        action.__code__.co_firstlineno,
        action.__code__.co_name,
    ]
    if suffix is not None:
        parts.append(suffix)
    return ':'.join(map(str, parts))


class Handler(object):
    """
    Class representing a reactive flag handler.
    """
    _HANDLERS = {}
    _CONSUMED_FLAGS = set()

    @classmethod
    def get(cls, action, suffix=None):
        """
        Get or register a handler for the given action.

        :param func action: Callback that is called when invoking the Handler
        :param func suffix: Optional suffix for the handler's ID
        """
        action_id = _action_id(action, suffix)
        if action_id not in cls._HANDLERS:
            if LOG_OPTS['register']:
                hookenv.log('Registering reactive handler for %s' % _short_action_id(action, suffix),
                            level=hookenv.DEBUG)
            cls._HANDLERS[action_id] = cls(action, suffix)
        return cls._HANDLERS[action_id]

    @classmethod
    def get_handlers(cls):
        """
        Get all registered handlers.
        """
        return cls._HANDLERS.values()

    @classmethod
    def clear(cls):
        """
        Clear all registered handlers.
        """
        cls._HANDLERS = {}

    def __init__(self, action, suffix=None):
        """
        Create a new Handler.

        :param func action: Callback that is called when invoking the Handler
        :param func suffix: Optional suffix for the handler's ID
        """
        self._action_id = _short_action_id(action, suffix)
        self._action = action
        self._args = []
        self._predicates = []
        self._post_callbacks = []
        self._flags = set()

    def id(self):
        return self._action_id

    def add_args(self, args):
        """
        Add arguments to be passed to the action when invoked.

        :param args: Any sequence or iterable, which will be lazily evaluated
            to provide args.  Subsequent calls to :meth:`add_args` can be used
            to add additional arguments.
        """
        self._args.append(args)

    @property
    def has_args(self):
        """
        Whether or not this Handler has had any args added via :meth:`add_args`.
        """
        return len(self._args) > 0

    def add_predicate(self, predicate):
        """
        Add a new predicate callback to this handler.
        """
        _predicate = predicate
        if isinstance(predicate, partial):
            _predicate = 'partial(%s, %s, %s)' % (predicate.func, predicate.args, predicate.keywords)
        if LOG_OPTS['register']:
            hookenv.log('  Adding predicate for %s: %s' % (self.id(), _predicate), level=hookenv.DEBUG)
        self._predicates.append(predicate)

    def add_post_callback(self, callback):
        """
        Add a callback to be run after the action is invoked.
        """
        self._post_callbacks.append(callback)

    def test(self):
        """
        Check the predicate(s) and return True if this handler should be invoked.
        """
        if self._flags and not FlagWatch.watch(self._action_id, self._flags):
            return False
        return all(predicate() for predicate in self._predicates)

    def _get_args(self):
        """
        Lazily evaluate the args.
        """
        if not hasattr(self, '_args_evaled'):
            # cache the args in case handler is re-invoked due to flags change
            self._args_evaled = list(chain.from_iterable(self._args))
        return self._args_evaled

    def invoke(self):
        """
        Invoke this handler.
        """
        args = self._get_args()
        self._action(*args)
        for callback in self._post_callbacks:
            callback()

    def register_flags(self, flags):
        """
        Register flags as being relevant to this handler.

        Relevant flags will be used to determine if the handler should
        be re-invoked due to changes in the set of active flags.  If this
        handler has already been invoked during this :func:`dispatch` run
        and none of its relevant flags have been set or removed since then,
        then the handler will be skipped.

        This is also used for linting and composition purposes, to determine
        if a layer has unhandled flags.
        """
        self._CONSUMED_FLAGS.update(flags)
        self._flags.update(flags)


class ExternalHandler(Handler):
    """
    A variant Handler for external executable actions (such as bash scripts).

    External handlers must adhere to the following protocol:

      * The handler can be any executable

      * When invoked with the ``--test`` command-line flag, it should exit with
        an exit code of zero to indicate that the handler should be invoked, and
        a non-zero exit code to indicate that it need not be invoked.  It can
        also provide a line of output to be passed to the ``--invoke`` call, e.g.,
        to indicate which sub-handlers should be invoked.  The handler should
        **not** perform its action when given this flag.

      * When invoked with the ``--invoke`` command-line flag (which will be
        followed by any output returned by the ``--test`` call), the handler
        should perform its action(s).
    """
    @classmethod
    def register(cls, filepath):
        if filepath not in Handler._HANDLERS:
            _filepath = os.path.relpath(filepath, hookenv.charm_dir())
            if LOG_OPTS['register']:
                hookenv.log('Registering external reactive handler for %s' % _filepath, level=hookenv.DEBUG)
            Handler._HANDLERS[filepath] = cls(filepath)
        return Handler._HANDLERS[filepath]

    def __init__(self, filepath):
        self._filepath = filepath
        self._test_output = ''

    def id(self):
        _filepath = os.path.relpath(self._filepath, hookenv.charm_dir())
        return '%s "%s"' % (_filepath, self._test_output)

    def test(self):
        """
        Call the external handler to test whether it should be invoked.
        """
        # flush to ensure external process can see flags as they currently
        # are, and write flags (flush releases lock)
        unitdata.kv().flush()
        try:
            proc = subprocess.Popen([self._filepath, '--test'], stdout=subprocess.PIPE, env=os.environ)
        except OSError as oserr:
            if oserr.errno == errno.ENOEXEC:
                raise BrokenHandlerException(self._filepath)
            raise
        self._test_output, _ = proc.communicate()
        return proc.returncode == 0

    def invoke(self):
        """
        Call the external handler to be invoked.
        """
        # flush to ensure external process can see flags as they currently
        # are, and write flags (flush releases lock)
        unitdata.kv().flush()
        subprocess.check_call([self._filepath, '--invoke', self._test_output], env=os.environ)


class FlagWatch(object):
    key = 'reactive.state_watch'

    @classmethod
    def _store(cls):
        return unitdata.kv()

    @classmethod
    def _get(cls):
        return cls._store().get(cls.key, {
            'iteration': 0,
            'changes': [],
            'pending': [],
        })

    @classmethod
    def _set(cls, data):
        cls._store().set(cls.key, data)

    @classmethod
    def reset(cls):
        cls._store().unset(cls.key)

    @classmethod
    def iteration(cls, i):
        data = cls._get()
        data['iteration'] = i
        cls._set(data)

    @classmethod
    def watch(cls, watcher, flags):
        data = cls._get()
        iteration = data['iteration']
        changed = bool(set(flags) & set(data['changes']))
        return iteration == 0 or changed

    @classmethod
    def change(cls, flag):
        data = cls._get()
        data['pending'].append(flag)
        cls._set(data)

    @classmethod
    def commit(cls):
        data = cls._get()
        data['changes'] = data['pending']
        data['pending'] = []
        cls._set(data)


def dispatch(restricted=False):
    """
    Dispatch registered handlers.

    When dispatching in restricted mode, only matching hook handlers are executed.

    Handlers are dispatched according to the following rules:

    * Handlers are repeatedly tested and invoked in iterations, until the system
      settles into quiescence (that is, until no new handlers match to be invoked).

    * In the first iteration, :func:`@hook <charms.reactive.decorators.hook>`
      and :func:`@action <charms.reactive.decorators.action>` handlers will
      be invoked, if they match.

    * In subsequent iterations, other handlers are invoked, if they match.

    * Added flags will not trigger new handlers until the next iteration,
      to ensure that chained flags are invoked in a predictable order.

    * Removed flags will cause the current set of matched handlers to be
      re-tested, to ensure that no handler is invoked after its matching
      flag has been removed.

    * Other than the guarantees mentioned above, the order in which matching
      handlers are invoked is undefined.

    * Flags are preserved between hook and action invocations, and all matching
      handlers are re-invoked for every hook and action.  There are
      :doc:`decorators <charms.reactive.decorators>` and
      :doc:`helpers <charms.reactive.helpers>`
      to prevent unnecessary reinvocations, such as
      :func:`~charms.reactive.decorators.only_once`.
    """
    FlagWatch.reset()

    def _test(to_test):
        return list(filter(lambda h: h.test(), to_test))

    def _invoke(to_invoke):
        while to_invoke:
            unitdata.kv().set('reactive.dispatch.removed_state', False)
            for handler in list(to_invoke):
                to_invoke.remove(handler)
                hookenv.log('Invoking reactive handler: %s' % handler.id(), level=hookenv.INFO)
                handler.invoke()
                if unitdata.kv().get('reactive.dispatch.removed_state'):
                    # re-test remaining handlers
                    to_invoke = _test(to_invoke)
                    break
        FlagWatch.commit()

    tracer().start_dispatch()

    # When in restricted context, only run hooks for that context.
    if restricted:
        unitdata.kv().set('reactive.dispatch.phase', 'restricted')
        hook_handlers = _test(Handler.get_handlers())
        tracer().start_dispatch_phase('restricted', hook_handlers)
        _invoke(hook_handlers)
        return

    unitdata.kv().set('reactive.dispatch.phase', 'hooks')
    hook_handlers = _test(Handler.get_handlers())
    tracer().start_dispatch_phase('hooks', hook_handlers)
    _invoke(hook_handlers)

    unitdata.kv().set('reactive.dispatch.phase', 'other')
    for i in range(100):
        FlagWatch.iteration(i)
        other_handlers = _test(Handler.get_handlers())
        if i == 0:
            tracer().start_dispatch_phase('other', other_handlers)
        tracer().start_dispatch_iteration(i, other_handlers)
        if not other_handlers:
            break
        _invoke(other_handlers)

    FlagWatch.reset()


def discover():
    """
    Discover handlers based on convention.

    Handlers will be loaded from the following directories and their subdirectories:

      * ``$CHARM_DIR/reactive/``
      * ``$CHARM_DIR/hooks/reactive/``
      * ``$CHARM_DIR/hooks/relations/``

    They can be Python files, in which case they will be imported and decorated
    functions registered.  Or they can be executables, in which case they must
    adhere to the :class:`ExternalHandler` protocol.
    """
    # Add $CHARM_DIR and $CHARM_DIR/hooks to sys.path so
    # 'import reactive.leadership', 'import relations.pgsql' works
    # as expected, as well as relative imports like 'import ..leadership'
    # or 'from . import leadership'. Without this, it becomes difficult
    # for layers to access APIs provided by other layers. This addition
    # needs to remain in effect, in case discovered modules are doing
    # late imports.
    _append_path(hookenv.charm_dir())
    _append_path(os.path.join(hookenv.charm_dir(), 'hooks'))

    for search_dir in ('reactive', 'hooks/reactive', 'hooks/relations'):
        search_path = os.path.join(hookenv.charm_dir(), search_dir)
        for dirpath, dirnames, filenames in os.walk(search_path):
            if os.path.basename(dirpath) == '__pycache__':
                continue
            for filename in filenames:
                filepath = os.path.join(dirpath, filename)
                _register_handlers_from_file(search_path, filepath)


def _append_path(d):
    if d not in sys.path:
        sys.path.append(d)


def _load_module(root, filepath):
    assert filepath.startswith(root + os.sep)
    assert filepath.endswith('.py')
    package = os.path.basename(root)  # 'reactive' or 'relations'
    assert package in ('reactive', 'relations')
    module = filepath[len(root):-3].replace(os.sep, '.')
    if module.endswith('.__init__'):
        module = module[:-9]

    # Standard import.
    return importlib.import_module(package + module)


def _is_external_handler(filepath):
    if not os.access(filepath, os.X_OK):
        return False
    _, ext = os.path.splitext(filepath)
    if ext not in ('', '.sh'):
        return False
    with open(filepath, 'rb') as fp:
        bytes = fp.read(1024)
    try:
        text = bytes.decode('utf8')
    except UnicodeDecodeError:
        # binary executable
        return True
    else:
        return text.startswith('#!')


def _register_handlers_from_file(root, filepath):
    if filepath.endswith('.py'):
        _load_module(root, filepath)
    elif _is_external_handler(filepath):
        ExternalHandler.register(filepath)



import sys
import weakref
from uuid import uuid4

from charms.reactive.flags import toggle_flag
from charms.reactive.endpoints import Endpoint


__all__ = [
    'Field',
    'BaseRequest',
    'BaseResponse',
    'RequesterEndpoint',
    'ResponderEndpoint',
]


class SetNameBackport(type):
    """
    Metaclass to backport the __set_name__ behavior for data descriptors,
    which was added in Python 3.6, to earlier versions.
    """
    if sys.version_info[0:2] < (3, 6):
        def __init__(cls, name, bases, namespace):
            super().__init__(name, bases, namespace)
            for attr_name in dir(cls):
                attr = getattr(cls, attr_name)
                if hasattr(attr, '__set_name__'):
                    attr.__set_name__(cls, attr_name)


class Field(property):
    """
    Defines a Field property for a Request or Response object.

    Can be set or retrieved like a normal attribute, and will be automatically
    serialized over the relation using JSON.
    """
    def __init__(self, description):
        self.__doc__ = description
        self._class = None
        self._name = None

    def __repr__(self):
        return '<Field {}.{}>'.format(self._class, self._name)

    def __set_name__(self, owner, name):
        self._class = owner.__name__
        self._name = name

    def __set__(self, instance, value):
        instance._update_field(self._name, value)

    def __get__(self, instance, owner):
        if instance is None:
            # called as class attribute, so return the Field instance instead
            # of blowing up on None not having a _get_field method (also allows
            # introspecting the class to find the fields)
            return self
        else:
            return instance._get_field(self._name)


class FieldHolderDictProxy(dict):
    """
    Base class for field holders that makes it act like a dict for easy
    serialization.
    """
    def __init__(self):
        cls = type(self)
        super().__init__({attr_name: getattr(self, attr_name)
                          for attr_name in dir(cls)
                          if isinstance(getattr(cls, attr_name), Field)})

    def _update_field(self, field_name, field_value):
        self[field_name] = field_value


class BaseRequest(FieldHolderDictProxy, metaclass=SetNameBackport):
    """
    Base class for requests using the request / response pattern.

    Subclasses **must** set the ``RESPONSE_CLASS`` attribute to a subclass of
    the :class:`BaseResponse` which defines the fields that the response will
    use.  They must also define additional attributes as :class:`Field`s.

    For example::

        class TLSResponse(BaseResponse):
            key = Field('Private key for the cert')
            cert = Field('Public cert info')


        class TLSRequest(BaseRequest):
            RESPONSE_CLASS = TLSResponse

            common_name = Field('Common Name (CN) for the cert to be created')
            sans = Field('List of Subject Alternative Names (SANs)')
    """
    RESPONSE_CLASS = None  # must be defined by subclass

    request_id = Field('UUID for this request.  Will be automatically generated.')

    _cache = None

    @classmethod
    def _load(cls, request_sources):
        """
        Load the requests and any associated responses from relation data.

        Must be called prior to requests being accessed.  This is done
        automatically by :class:`RequesterEndpoint` and
        :class:`ResponderEndpoint` base classes.

        :param request_sources: Where the requests should be loaded from.
          For example, the Endpoint initiating requests would use
          ``self.relations`` while the Endpoint processing requests might use
          ``self.all_joined_units``.
        """
        cls._cache = {}  # set here so that each subclass gets its own cache
        for source in request_sources:
            requests = (source.to_publish if hasattr(source, 'to_publish')
                        else source.received)
            for key, request_data in requests.items():
                if not key.startswith('request_'):
                    continue
                request = cls(source, request_data['request_id'])
                request.response = cls.RESPONSE_CLASS._load(request)
                cls._cache[request.request_id] = request

    @classmethod
    def create(cls, relation, **fields):
        """
        Create a new request.

        Fields and their values can be passed in to pre-populate the request as
        keyword arguments, or can be set individually on the resulting request.
        """
        request_id = fields.setdefault('request_id', str(uuid4()))
        # pre-populate the field data directly in the data store (more
        # efficient than calling _update_field for every field)
        relation.to_publish['request_' + request_id] = fields
        cls._cache[request_id] = cls(relation, request_id)
        return cls._cache[request_id]

    @classmethod
    def get(cls, request_id):
        """
        Get a specific request by ID.
        """
        return cls._cache.get(request_id)

    @classmethod
    def get_all(cls):
        """
        Get a list of all requests (in order of their ID).
        """
        return sorted(cls._cache.values(), key=lambda r: r._id)

    @classmethod
    def find(cls, relation=None, **fields):
        """
        Find the first request whose fields match the given values.

        :param relation: If given, look for the request on a specific relation.
        :param **fields: Name / value pairs to match by.
        """
        for request in cls._cache.values():
            if all(getattr(request, field_name) == field_value
                   for field_name, field_value in fields.items()):
                return request
        else:
            return None

    @classmethod
    def find_all(cls, **fields):
        """
        Find all requests whose fields match the given values.

        :param **fields: Name / value pairs to match by.
        """
        found = []
        for request in cls._cache.values():
            if all(getattr(request, field_name) == field_value
                   for field_name, field_value in fields.items()):
                found.append(request)
        return found

    @classmethod
    def create_or_update(cls, match_fields, relation, **fields):
        """
        Find a request and update it, or create a new one.

        If multiple requests match, only the first one is updated.

        :param match_fields: List of the field names to match by.
        :param relation: Relation to find or create the request on.
        :param **fields: Name / value pairs to match by and update to.
        :returns: The new or updated request.

        Example::

            for relation in self.relations:
                JobRequest.create_or_update(match_fields=['job_name'],
                                            relation=relation,
                                            job_name='foo',
                                            job_data=job_data)
        """
        request = cls.find(relation, **{field: fields[field]
                                        for field in match_fields})
        if request:
            # update
            for field_name, field_value in fields.items():
                request._update_field(field_name, field_value)
        else:
            # create
            request = cls.create(relation, **fields)
        return request

    def __init__(self, source, request_id):
        if self.RESPONSE_CLASS is None:
            raise TypeError('RESPONSE_CLASS must be defined by subclass')
        self._id = request_id  # cache the ID so that we can determine the key
        self._source = source
        self.response = None
        if self.is_received:
            self._source_data = source.received
        else:
            self._source_data = source.to_publish
        super().__init__()

    def __repr__(self):
        return '<{} {}>'.format(type(self).__name__, self._id)

    @property
    def _key(self):
        return 'request_{}'.format(self._id)

    @property
    def is_created(self):
        """
        Whether this request was created by this side of the relation.
        """
        return hasattr(self._source, 'to_publish')

    @property
    def is_received(self):
        """
        Whether this request was received by the other side of the relation.
        """
        return not self.is_created

    def create_response(self, **fields):
        """
        Create a response to this request.

        Fields and their values can be passed in to pre-populate the response
        as keyword arguments, or can be set individually on the resulting
        response.

        Returns the response object (which can also be accessed as
        ``request.response``).
        """
        self.response = self.RESPONSE_CLASS.create(self, **fields)
        return self.response

    def respond(self, **fields):
        """
        Respond to this request.  (Alias of :meth:`create_response`.)

        Fields and their values can be passed in to pre-populate the response
        as keyword arguments, or can be set individually on the resulting
        response.

        Returns the response object (which can also be accessed as
        ``request.response``).
        """
        return self.create_response(**fields)

    def _get_field(self, name):
        return self._source_data.get(self._key, {}).get(name)

    def _update_field(self, name, value):
        if self.is_received:
            raise AttributeError("can't change field for received request")
        if name == 'request_id':
            raise AttributeError("request_id can't be modified")
        data = self._source_data[self._key]
        data[name] = value
        # reading serialized dict gets a copy, so we have to explicitly write
        # it back to ensure updated items get written
        self._source_data[self._key] = data
        # update the proxy for serialization
        super()._update_field(name, value)

    @property
    def ingress_address(self):
        """
        Address to use if a connection to the requester is required.
        """
        return self._source_data['ingress-address']

    @property
    def egress_subnets(self):
        """
        Subnets over which network traffic to the requester will flow.
        """
        return self._source_data['egress-subnets']


class BaseResponse(FieldHolderDictProxy, metaclass=SetNameBackport):
    """
    Base class for responses using the request / response pattern.
    """
    @classmethod
    def _load(cls, request):
        response = cls(request)
        if response._key not in response._source_data:
            return None  # no response found
        else:
            return response

    @classmethod
    def create(cls, request, **fields):
        """
        Create a response to the given request.

        Fields and their values can be passed in to pre-populate the response
        as keyword arguments, or can be set individually on the resulting
        response.
        """
        if request.is_created:
            raise ValueError("can't respond to requests we created")
        # pre-populate the field data directly in the data store (more
        # efficient than calling _update_field for every field)
        request._source.relation.to_publish['response_' + request.request_id] = fields
        response = cls(request)
        request.response = response
        return response

    def __init__(self, request):
        self.request = weakref.proxy(request)
        if request.is_received:
            self._source_data = request._source.relation.to_publish
        else:
            for unit in request._source.joined_units:
                if self._key in unit.received:
                    self._source_data = unit.received
                    break
            else:
                self._source_data = {}
        super().__init__()

    def __repr__(self):
        return '<{} {}>'.format(type(self).__name__, self.request._id)

    @property
    def _key(self):
        return 'response_{}'.format(self.request.request_id)

    @property
    def is_received(self):
        return not self.request.is_received

    def _get_field(self, name):
        return self._source_data.get(self._key, {}).get(name)

    def _update_field(self, name, value):
        if self.is_received:
            raise AttributeError("can't change field for received response")
        data = self._source_data[self._key]
        data[name] = value
        # reading serialized dict gets a copy, so we have to explicitly write
        # it back to ensure updated items get written
        self._source_data[self._key] = data
        # update the proxy for serialization
        super()._update_field(name, value)


class FieldFinders(type):
    """
    Metaclass for defining ``response_by_FIELD`` methods on
    ``RequesterEndpoint`` classes.

    This is done as a metaclass to handle fields on derived classes but
    to still allow the methods to show up in the class's documentation.
    """
    # Is this too much magic?  Should we just provide response_by_field
    # and leave it up to the endpoint author to define which search methods
    # make sense?
    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)
        for field_name in dir(cls.REQUEST_CLASS):
            field = getattr(cls.REQUEST_CLASS, field_name)
            if not isinstance(field, Field):
                continue

            def _make_finder(field_name):
                # has to be nested to make closure work properly
                def _finder(self, field_value, relation=None):
                    return self.response_by_field(**{field_name: field_value,
                                                     'relation': relation})
                _finder.__name__ = 'response_by_{}'.format(field_name)
                _finder.__doc__ = """
                    Find the response by the value of its request's {} field.

                    If ``relation`` is given, limit the search to that
                    relation.
                    """.format(field_name)
                return _finder
            _finder = _make_finder(field_name)
            setattr(cls, _finder.__name__, _finder)


class RequesterEndpoint(Endpoint, metaclass=FieldFinders):
    """
    Base class for Endpoints that create requests in the request / response
    pattern.

    Subclasses **must** set the ``REQUEST_CLASS`` attribute to a subclass
    of :class:`BaseRequest` which defines the fields the request will use.

    Will automatically manage the following flags:

      * ``endpoint.{endpoint_name}.has_responses`` Set if any responses are
        available
      * ``endpoint.{endpoint_name}.all_responses`` Set if all requests have
        responses.
    """
    REQUEST_CLASS = None  # must be defined by subclass

    def __init__(self, *args, **kwargs):
        if self.REQUEST_CLASS is None:
            raise TypeError('REQUEST_CLASS must be defined by subclass')
        super().__init__(*args, **kwargs)
        self.REQUEST_CLASS._load(self.relations)

    def _manage_flags(self):
        super()._manage_flags()
        toggle_flag(self.expand_name('endpoint.{endpoint_name}.has_responses'),
                    self.responses)
        toggle_flag(self.expand_name('endpoint.{endpoint_name}.all_responses'),
                    len(self.responses) == len(self.requests))

    @property
    def requests(self):
        """
        A list of all requests which have been submitted.
        """
        return self.REQUEST_CLASS.get_all()

    @property
    def responses(self):
        """
        A list of all responses which have been received.
        """
        return [request.response
                for request in self.requests
                if request.response is not None]

    def response_by_field(self, relation=None, **fields):
        """
        Find a response by the value(s) of fields on its request.

        :param relation: If given, limit the search to that relation.
        :param **fields: Name / value pairs to match by.
        """
        request = self.REQUEST_CLASS.find(relation=relation, **fields)
        return request.response if request else None


class ResponderEndpoint(Endpoint):
    """
    Base class for Endpoints that respond to requests in the request / response
    pattern.

    Subclasses **must** set the ``REQUEST_CLASS`` attribute to a subclass
    of :class:`BaseRequest` which defines the fields the request will use.

    Will automatically manage the following flags:

      * ``endpoint.{endpoint_name}.has_requests`` Set if any requests are
        available
      * ``endpoint.{endpoint_name}.new_requests`` Set if any unhandled requests
        are available.
    """
    REQUEST_CLASS = None  # must be defined by subclass

    def __init__(self, *args, **kwargs):
        if self.REQUEST_CLASS is None:
            raise TypeError('REQUEST_CLASS must be defined by subclass')
        super().__init__(*args, **kwargs)
        self.REQUEST_CLASS._load(self.all_joined_units)

    def _manage_flags(self):
        super()._manage_flags()
        toggle_flag(self.expand_name('endpoint.{endpoint_name}.has_requests'),
                    self.all_requests)
        toggle_flag(self.expand_name('endpoint.{endpoint_name}.new_requests'),
                    self.new_requests)

    @property
    def all_requests(self):
        """
        A list of all requests, including ones which have been responded to.
        """
        return self.REQUEST_CLASS.get_all()

    @property
    def new_requests(self):
        """
        A list of requests which have not been responded.

        Requests should be handled by the charm and then responded to by
        calling ``request.respond(...)``.
        """
        return list(filter(lambda request: request.response is None,
                           self.all_requests))



# Copyright 2014-2017 Canonical Limited.
#
# This file is part of charms.reactive.
#
# charms.reactive is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charm-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import types
import mock
import unittest

from charmhelpers.core import hookenv
from charms.reactive import relations


class DummyRelationSubclass(relations.RelationBase):
    auto_accessors = ['field-one', 'field-two']


class TestFactory(unittest.TestCase):
    @mock.patch.object(relations.hookenv, 'log')
    @mock.patch.object(relations, 'importlib')
    @mock.patch.object(relations.hookenv, 'charm_dir')
    @mock.patch.object(relations.hookenv, 'relation_to_role_and_interface')
    def test_relation_factory_import_fail(self, relation_to_role_and_interface,
                                          charm_dir, importlib, log):
        relation_to_role_and_interface.return_value = ('role', 'interface')
        charm_dir.return_value = 'charm_dir'
        importlib.import_module.side_effect = ImportError
        self.assertIsNone(relations.relation_factory('relname'))
        relation_to_role_and_interface.assert_called_once_with('relname')
        self.assertEqual(importlib.import_module.call_args_list, [
            mock.call('reactive.relations.interface.role'),
            mock.call('relations.interface.role'),
        ])
        log.assert_called_once_with(mock.ANY, relations.hookenv.ERROR)

    @mock.patch.object(relations, 'entry_points')
    @mock.patch.object(relations, '_find_relation_factory')
    @mock.patch.object(relations.hookenv, 'log')
    @mock.patch.object(relations, '_relation_module')
    @mock.patch.object(relations.hookenv, 'charm_dir')
    @mock.patch.object(relations.hookenv, 'relation_to_role_and_interface')
    def test_relation_factory(self, relation_to_role_and_interface, charm_dir,
                              rel_mod, log, find_factory, entry_points):
        relation_to_role_and_interface.return_value = ('role', 'interface')
        charm_dir.return_value = 'charm_dir'
        rel_mod.return_value = 'module'
        find_factory.return_value = 'fact'

        class MockRelFactory:
            @classmethod
            def from_name(cls, name):
                if name == 'foo':
                    return mock.Mock()

            load = mock.Mock()

        entry_points.return_value = {
            'charms.reactive.relation_factory': [mock.Mock(load=lambda: MockRelFactory)],
        }
        relations.RelationFactory.discover()
        assert MockRelFactory.load.called

        self.assertEqual(relations.relation_factory('relname'), 'fact')
        relation_to_role_and_interface.assert_called_once_with('relname')
        rel_mod.assert_called_once_with('role', 'interface')
        find_factory.assert_called_once_with('module')
        self.assertIs(relations.RelationFactory.get_factory('foo'), MockRelFactory)

    def test_find_relation_factory(self):
        mod = types.ModuleType('mod')
        mod.__name__ = 'here'

        # Noise to be ignored
        mod.RelationFactory = relations.RelationFactory
        mod.RelationBase = relations.RelationBase

        # One subclass of RelationFactory
        class Rel(relations.RelationFactory):
            pass
        mod.Rel = Rel
        self.assertIs(relations._find_relation_factory(mod), Rel)

        # One subclass of RelationBase
        class Rel2(relations.RelationBase):
            pass
        del mod.Rel
        mod.Rel2 = Rel2
        self.assertIs(relations._find_relation_factory(mod), Rel2)

        # Subclass, and a subclass of it.
        class Rel3(Rel2):
            pass
        mod.Rel3 = Rel3
        self.assertIs(relations._find_relation_factory(mod), Rel3)

        # A 2nd valid choice
        mod.Rel4 = Rel3
        with self.assertRaises(RuntimeError):
            relations._find_relation_factory(mod)

    def test_endpoint_from_name_missing_name(self):
        self.assertIsNone(relations.endpoint_from_name(None))

    @mock.patch.object(relations, 'relation_factory')
    def test_endpoint_from_name_missing_factory(self, relation_factory):
        relation_factory.return_value = None
        self.assertIsNone(relations.endpoint_from_name('relname'))
        relation_factory.assert_called_once_with('relname')

    @mock.patch.object(relations, 'relation_factory')
    def test_endpoint_from_name(self, relation_factory):
        relation_factory('relname').from_name.return_value = 'relinstance'
        relation_factory.reset_mock()

        self.assertEqual(relations.endpoint_from_name('relname'),
                         'relinstance')

        relation_factory.assert_called_once_with('relname')
        relation_factory('relname').from_name.assert_called_once_with('relname')


class TestAutoAccessors(unittest.TestCase):
    def setUp(self):
        kv_p = mock.patch.object(relations.unitdata, 'kv')
        self.kv = kv_p.start()
        self.addCleanup(kv_p.stop)
        self.kv.return_value.get.side_effect = lambda k, v=None: v

    def test_accessor_doc(self):
        self.assertEqual(DummyRelationSubclass.field_one.__doc__, 'Get the field-one, if available, or None.')
        self.assertEqual(DummyRelationSubclass.field_one.__name__, 'field_one')
        self.assertEqual(DummyRelationSubclass.field_one.__module__, 'test_relations')

    def test_accessor(self):
        rel = DummyRelationSubclass('rel', [])
        rel.get_remote = mock.Mock(side_effect=['value1', 'value2'])
        self.assertEqual(rel.field_one(), 'value1')
        self.assertEqual(rel.field_two(), 'value2')
        self.assertEqual(rel.get_remote.call_args_list, [
            mock.call('field-one'),
            mock.call('field-two'),
        ])


class TestRelationBase(unittest.TestCase):
    def setUp(self):
        hookenv.cache = {}

    def test_find_subclass(self):
        self.assertIsNone(relations.RelationBase._find_subclass(unittest))
        self.assertIsNone(relations.RelationBase._find_subclass(relations))  # want strictly subclasses
        self.assertIs(
            relations.RelationBase._find_subclass(sys.modules[__name__]),
            DummyRelationSubclass)

    @mock.patch.object(relations.RelationBase, 'from_name')
    @mock.patch.object(relations.Conversation, 'load')
    @mock.patch.object(relations, '_get_flag_value')
    def test_from_state(self, get_flag_value, load, from_name):
        load.return_value = 'conv.load'
        get_flag_value.side_effect = [{'relation': 'relname', 'conversations': ['conv']}, None]
        from_name.side_effect = lambda rn, c: 'from_name(%s, %s)' % (rn, c)
        self.assertEqual(relations.RelationBase.from_flag('state'), 'from_name(relname, conv.load)')
        self.assertEqual(relations.RelationBase.from_flag('no-state'), None)

    @mock.patch.object(relations.hookenv, 'relation_to_role_and_interface')
    @mock.patch.object(relations.Conversation, 'join')
    @mock.patch.object(relations.RelationBase, '_find_impl')
    def test_from_name(self, _find_impl, join, relation_to_role_and_interface):
        relation_to_role_and_interface.return_value = ('role', 'interface')
        r1 = mock.Mock(name='R1')
        _find_impl.side_effect = [None, r1, None]
        join.return_value = 'conv.join'

        self.assertIs(relations.RelationBase.from_name('none'), None)
        self.assertIs(relations.RelationBase.from_name('rel'), r1())

        # test cache
        self.assertEqual(_find_impl.call_count, 2)
        self.assertIs(relations.RelationBase.from_name('rel'), r1())
        self.assertEqual(_find_impl.call_count, 2)

        # test instantation
        relations.RelationBase._cache.clear()
        _find_impl.side_effect = [relations.RelationBase]
        res = relations.RelationBase.from_name('rel')
        self.assertIsInstance(res, relations.RelationBase)
        self.assertEqual(res._relation_name, 'rel')
        self.assertEqual(res.relation_name, 'rel')
        self.assertEqual(res._conversations, ['conv.join'])
        self.assertEqual(res.conversations(), ['conv.join'])

    @mock.patch.dict('sys.modules')
    @mock.patch.object(relations.Conversation, 'join')
    @mock.patch.object(relations, 'hookenv')
    def test_cold_import(self, hookenv, conv_join):
        tests_dir = os.path.dirname(__file__)
        hookenv.charm_dir.return_value = os.path.join(tests_dir, 'data')
        sys.modules.pop('relations.hyphen-ated.peer', None)
        hookenv.relation_to_role_and_interface.return_value = ('peer',
                                                               'hyphen-ated')
        relations.RelationBase._cache.clear()
        assert relations.RelationBase.from_name('test') is not None

    @mock.patch.dict('sys.modules')
    @mock.patch.object(relations, 'hookenv')
    @mock.patch.object(relations.RelationBase, '_find_subclass')
    def test_find_impl(self, find_subclass, hookenv):
        tests_dir = os.path.dirname(__file__)
        hookenv.charm_dir.return_value = os.path.join(tests_dir, 'data')
        self.assertIsNone(relations.RelationBase._find_impl('role',
                                                            'interface'))
        assert not find_subclass.called

        sys.modules['relations.interface.role'] = mock.sentinel.m1
        find_subclass.return_value = mock.sentinel.r1
        self.assertIs(relations.RelationBase._find_impl('role', 'interface'),
                      mock.sentinel.r1)
        find_subclass.assert_called_once_with(mock.sentinel.m1)

    @mock.patch.object(relations, 'hookenv')
    def test_conversation(self, hookenv):
        hookenv.remote_unit.return_value = 'remote_unit'
        hookenv.remote_service_name.return_value = 'remote_service_name'

        conv1 = mock.Mock(scope='remote_unit')
        conv2 = mock.Mock(scope='remote_service_name')
        conv3 = mock.Mock(scope='global')
        conv4 = mock.Mock(scope='explicit')
        rel = DummyRelationSubclass('relname', [conv1, conv2, conv3, conv4])

        self.assertIs(rel.conversation(), conv1)
        rel.scope = relations.scopes.SERVICE
        self.assertIs(rel.conversation(), conv2)
        rel.scope = relations.scopes.GLOBAL
        self.assertIs(rel.conversation(), conv3)
        self.assertIs(rel.conversation('explicit'), conv4)
        self.assertRaises(ValueError, rel.conversation, 'none')
        rel.scope = None
        self.assertRaises(ValueError, rel.conversation)

    def test_set_state(self):
        conv = mock.Mock(name='conv')
        rb = relations.RelationBase('relname', 'unit')
        rb.conversation = mock.Mock(return_value=conv)
        rb.set_state('state', 'scope')
        rb.conversation.assert_called_once_with('scope')
        conv.set_state.assert_called_once_with('state')
        rb.conversation.reset_mock()
        conv.reset_mock()
        rb.set_flag('state', 'scope')
        rb.conversation.assert_called_once_with('scope')
        conv.set_state.assert_called_once_with('state')

    def test_remove_state(self):
        conv = mock.Mock(name='conv')
        rb = relations.RelationBase('relname', 'unit')
        rb.conversation = mock.Mock(return_value=conv)
        rb.remove_state('state', 'scope')
        rb.conversation.assert_called_once_with('scope')
        conv.remove_state.assert_called_once_with('state')
        rb.conversation.reset_mock()
        conv.reset_mock()
        rb.remove_flag('state', 'scope')
        rb.conversation.assert_called_once_with('scope')
        conv.remove_state.assert_called_once_with('state')

    def test_is_state(self):
        conv = mock.Mock(name='conv')
        rb = relations.RelationBase('relname', 'unit')
        rb.conversation = mock.Mock(return_value=conv)
        rb.conversation.return_value.is_state.return_value = False
        assert not rb.is_state('state', 'scope')
        rb.conversation.assert_called_once_with('scope')
        conv.is_state.assert_called_once_with('state')
        rb.conversation.return_value.is_state.return_value = True
        assert rb.is_state('state', 'scope')
        assert rb.is_flag_set('state', 'scope')

    def test_toggle_state(self):
        conv = mock.Mock(name='conv')
        rb = relations.RelationBase('relname', 'unit')
        rb.conversation = mock.Mock(return_value=conv)
        rb.toggle_state('state', 'active', 'scope')
        rb.conversation.assert_called_once_with('scope')
        conv.toggle_state.assert_called_once_with('state', 'active')

        conv.toggle_state.reset_mock()
        rb.toggle_state('state')
        conv.toggle_state.assert_called_once_with('state',
                                                  relations.TOGGLE)

        conv.toggle_state.reset_mock()
        rb.toggle_flag('state')
        conv.toggle_state.assert_called_once_with('state',
                                                  relations.TOGGLE)

    def test_set_remote(self):
        conv = mock.Mock(name='conv')
        rb = relations.RelationBase('relname', 'unit')
        rb.conversation = mock.Mock(return_value=conv)
        rb.set_remote('key', 'value', data='data', scope='scope', kwarg='value')
        rb.conversation.assert_called_once_with('scope')
        conv.set_remote.assert_called_once_with('key', 'value', 'data', kwarg='value')

    def test_get_remote(self):
        conv = mock.Mock(name='conv')
        rb = relations.RelationBase('relname', 'unit')
        rb.conversation = mock.Mock(return_value=conv)
        rb.get_remote('key', 'default', 'scope')
        rb.conversation.assert_called_once_with('scope')
        conv.get_remote.assert_called_once_with('key', 'default')

    def test_set_local(self):
        conv = mock.Mock(name='conv')
        rb = relations.RelationBase('relname', 'unit')
        rb.conversation = mock.Mock(return_value=conv)
        rb.set_local('key', 'value', data='data', scope='scope', kwarg='value')
        rb.conversation.assert_called_once_with('scope')
        conv.set_local.assert_called_once_with('key', 'value', 'data', kwarg='value')

    def test_get_local(self):
        conv = mock.Mock(name='conv')
        rb = relations.RelationBase('relname', 'unit')
        rb.conversation = mock.Mock(return_value=conv)
        rb.get_local('key', 'default', 'scope')
        rb.conversation.assert_called_once_with('scope')
        conv.get_local.assert_called_once_with('key', 'default')


class TestConversation(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not hasattr(cls, 'assertItemsEqual'):
            cls.assertItemsEqual = cls.assertCountEqual

    def test_key(self):
        c1 = relations.Conversation('rel', ['unit'], 'scope')
        self.assertEqual(c1.key, 'reactive.conversations.rel.scope')

    @mock.patch.object(relations.hookenv, 'relation_ids')
    def test_relation_ids(self, relation_ids):
        relation_ids.return_value = ['rel:1', 'rel:2']
        c1 = relations.Conversation('rel:0', [], 'scope')
        self.assertEqual(c1.relation_ids, ['rel:0'])
        assert not relation_ids.called

        c2 = relations.Conversation('rel', [], relations.scopes.GLOBAL)
        self.assertEqual(c2.relation_ids, ['rel:1', 'rel:2'])
        relation_ids.assert_called_once_with('rel')

    @mock.patch.object(relations, 'unitdata')
    @mock.patch.object(relations, 'hookenv')
    def test_join(self, hookenv, unitdata):
        hookenv.relation_type.return_value = 'relation_type'
        hookenv.relation_id.return_value = 'relation_type:0'
        hookenv.remote_unit.return_value = 'service/0'
        hookenv.remote_service_name.return_value = 'service'
        unitdata.kv().get.side_effect = [
            {
                'namespace': 'relation_type',
                'units': [],
                'scope': 'my-global',
            },
            {
                'namespace': 'relation_type:0',
                'units': ['service/1'],
                'scope': 'service',
            },
            {
                'namespace': 'relation_type:0',
                'units': [],
                'scope': 'service/0',
            },
        ]
        conv = relations.Conversation.join('my-global')
        self.assertEqual(conv.namespace, 'relation_type')
        self.assertEqual(conv.units, {'service/0'})
        self.assertEqual(conv.scope, 'my-global')
        unitdata.kv().get.assert_called_with('reactive.conversations.relation_type.my-global', {
            'namespace': 'relation_type',
            'scope': 'my-global',
            'units': [],
        })
        unitdata.kv().set.assert_called_with('reactive.conversations.relation_type.my-global', {
            'namespace': 'relation_type',
            'units': ['service/0'],
            'scope': 'my-global',
        })

        conv = relations.Conversation.join(relations.scopes.SERVICE)
        self.assertEqual(conv.namespace, 'relation_type:0')
        self.assertEqual(conv.units, {'service/0', 'service/1'})
        self.assertEqual(conv.scope, 'service')
        unitdata.kv().get.assert_called_with('reactive.conversations.relation_type:0.service', {
            'namespace': 'relation_type:0',
            'scope': 'service',
            'units': [],
        })
        unitdata.kv().set.assert_called_with('reactive.conversations.relation_type:0.service', {
            'namespace': 'relation_type:0',
            'units': ['service/0', 'service/1'],
            'scope': 'service',
        })

        conv = relations.Conversation.join(relations.scopes.UNIT)
        self.assertEqual(conv.relation_name, 'relation_type')
        self.assertEqual(conv.units, {'service/0'})
        self.assertEqual(conv.scope, 'service/0')
        unitdata.kv().get.assert_called_with('reactive.conversations.relation_type:0.service/0', {
            'namespace': 'relation_type:0',
            'scope': 'service/0',
            'units': [],
        })
        unitdata.kv().set.assert_called_with('reactive.conversations.relation_type:0.service/0', {
            'namespace': 'relation_type:0',
            'units': ['service/0'],
            'scope': 'service/0',
        })

    @mock.patch.object(relations, 'unitdata')
    @mock.patch.object(relations, 'hookenv')
    def test_depart(self, hookenv, unitdata):
        hookenv.remote_unit.return_value = 'service/0'
        conv = relations.Conversation('rel', ['service/0', 'service/1'], 'scope')
        conv.depart()
        self.assertEqual(conv.units, {'service/1'}, 'scope')
        unitdata.kv().set.assert_called_with(conv.key, {
            'namespace': 'rel',
            'units': ['service/1'],
            'scope': 'scope',
        })
        assert not unitdata.kv().unset.called

        unitdata.kv().set.reset_mock()
        hookenv.remote_unit.return_value = 'service/1'
        conv.depart()
        assert not unitdata.kv().set.called
        unitdata.kv().unset.assert_called_with(conv.key)

    @mock.patch.object(relations, 'unitdata')
    def test_load(self, unitdata):
        unitdata.kv().get.side_effect = [
            {'namespace': 'rel:1', 'units': ['service/0'], 'scope': 'scope'},
            None,
            {'namespace': 'rel:2', 'units': ['service/1'], 'scope': 'service'},
        ]
        convs = relations.Conversation.load(['key1', 'key2', 'key3'])
        self.assertEqual(len(convs), 2)
        self.assertEqual(convs[0].relation_name, 'rel')
        self.assertEqual(convs[0].units, {'service/0'})
        self.assertEqual(convs[0].scope, 'scope')
        self.assertEqual(convs[1].relation_name, 'rel')
        self.assertEqual(convs[1].units, {'service/1'})
        self.assertEqual(convs[1].scope, 'service')
        self.assertEqual(unitdata.kv().get.call_args_list, [
            mock.call('key1'), mock.call('key2'), mock.call('key3'),
        ])

    @mock.patch.object(relations, 'set_flag')
    @mock.patch.object(relations, '_get_flag_value')
    def test_set_state(self, get_flag_value, set_flag):
        conv = relations.Conversation('rel', ['service/0', 'service/1'], 'scope')
        get_flag_value.return_value = {'conversations': ['foo']}
        conv.set_state('{relation_name}.bar')
        set_flag.assert_called_once_with('rel.bar', {'conversations': ['foo', 'reactive.conversations.rel.scope']})
        get_flag_value.assert_called_once_with('rel.bar', {'relation': 'rel', 'conversations': []})
        conv.set_state('{relation_name}.bar')
        self.assertEqual(set_flag.call_count, 2)
        set_flag.assert_called_with('rel.bar', {'conversations': ['foo', 'reactive.conversations.rel.scope']})
        conv.set_flag('{relation_name}.qux')
        self.assertEqual(set_flag.call_count, 3)
        set_flag.assert_called_with('rel.qux', {'conversations': ['foo', 'reactive.conversations.rel.scope']})

    @mock.patch.object(relations, 'clear_flag')
    @mock.patch.object(relations, 'set_flag')
    @mock.patch.object(relations, '_get_flag_value')
    def test_remove_state(self, get_flag_value, set_flag, clear_flag):
        conv = relations.Conversation('rel', ['service/0', 'service/1'], 'scope')
        get_flag_value.side_effect = [
            None,
            {'conversations': ['foo', 'reactive.conversations.rel.scope']},
            {'conversations': ['reactive.conversations.rel.scope']},
            {'conversations': ['reactive.conversations.rel.scope']},
        ]

        conv.remove_state('{relation_name}.bar')
        get_flag_value.assert_called_once_with('rel.bar')
        assert not set_flag.called
        assert not clear_flag.called

        conv.remove_state('{relation_name}.bar')
        set_flag.assert_called_once_with('rel.bar', {'conversations': ['foo']})
        assert not clear_flag.called

        set_flag.reset_mock()
        conv.remove_state('{relation_name}.bar')
        assert not set_flag.called
        clear_flag.assert_called_once_with('rel.bar')

        clear_flag.reset_mock()
        conv.remove_flag('{relation_name}.bar')
        assert not set_flag.called
        clear_flag.assert_called_once_with('rel.bar')

    @mock.patch.object(relations, '_get_flag_value')
    def test_is_state(self, get_flag_value):
        conv = relations.Conversation('rel', ['service/0', 'service/1'], 'scope')
        get_flag_value.side_effect = [
            None,
            {'conversations': ['foo']},
            {'conversations': ['reactive.conversations.rel.scope']},

            None,
            {'conversations': ['foo']},
            {'conversations': ['reactive.conversations.rel.scope']},
        ]

        assert not conv.is_state('{relation_name}.bar')
        assert not conv.is_state('{relation_name}.bar')
        assert conv.is_state('{relation_name}.bar')

        assert not conv.is_flag_set('{relation_name}.bar')
        assert not conv.is_flag_set('{relation_name}.bar')
        assert conv.is_flag_set('{relation_name}.bar')

    def test_toggle_state(self):
        conv = relations.Conversation('rel', ['service/0', 'service/1'], 'scope')
        conv.is_state = mock.Mock(side_effect=[True, False])
        conv.set_state = mock.Mock()
        conv.remove_state = mock.Mock()

        conv.toggle_state('foo')
        self.assertEqual(conv.remove_state.call_count, 1)
        conv.toggle_state('foo', None)
        self.assertEqual(conv.remove_state.call_count, 2)
        conv.toggle_state('foo')
        self.assertEqual(conv.set_state.call_count, 1)
        conv.toggle_state('foo', True)
        self.assertEqual(conv.set_state.call_count, 2)
        conv.toggle_state('foo', False)
        self.assertEqual(conv.remove_state.call_count, 3)
        conv.toggle_flag('foo', True)
        self.assertEqual(conv.set_state.call_count, 3)

    @mock.patch.object(relations.hookenv, 'relation_set')
    @mock.patch.object(relations.Conversation, 'relation_ids', ['rel:1', 'rel:2'])
    def test_set_remote(self, relation_set):
        conv = relations.Conversation('rel', ['service/0', 'service/1'], 'scope')

        conv.set_remote('foo', 'bar')
        self.assertEqual(relation_set.call_args_list, [
            mock.call('rel:1', {'foo': 'bar'}),
            mock.call('rel:2', {'foo': 'bar'}),
        ])
        relation_set.reset_mock()

        conv.set_remote(data={'foo': 'bar'})
        self.assertEqual(relation_set.call_args_list, [
            mock.call('rel:1', {'foo': 'bar'}),
            mock.call('rel:2', {'foo': 'bar'}),
        ])
        relation_set.reset_mock()

        conv.set_remote(foo='bar')
        self.assertEqual(relation_set.call_args_list, [
            mock.call('rel:1', {'foo': 'bar'}),
            mock.call('rel:2', {'foo': 'bar'}),
        ])
        relation_set.reset_mock()

        conv.set_remote('foo', 'bof', {'foo': 'bar'}, foo='qux')
        self.assertEqual(relation_set.call_args_list, [
            mock.call('rel:1', {'foo': 'qux'}),
            mock.call('rel:2', {'foo': 'qux'}),
        ])

        relation_set.reset_mock()
        conv.set_remote()
        assert not relation_set.called

    @mock.patch.object(relations.hookenv, 'relation_get')
    @mock.patch.object(relations.hookenv, 'related_units')
    @mock.patch.object(relations.Conversation, 'relation_ids', ['rel:1',
                                                                'rel:2'])
    def test_get_remote(self, related_units, relation_get):
        conv = relations.Conversation('rel',
                                      ['srv1/0', 'srv2/0', 'srv2/1'],
                                      'scope')

        # set on at least one remote
        related_units.side_effect = [['srv1/0', 'srv1/1'], ['srv2/1']]
        relation_get.side_effect = [None, 'value']
        self.assertEqual(conv.get_remote('key', 'default'), 'value')
        self.assertEqual(related_units.call_args_list, [mock.call('rel:1'),
                                                        mock.call('rel:2')])
        self.assertEqual(relation_get.call_args_list, [
            mock.call('key', 'srv1/0', 'rel:1'),
            mock.call('key', 'srv2/1', 'rel:2'),
        ])

        # not set on any remote
        related_units.side_effect = [[], ['srv2/0']]
        relation_get.side_effect = [None]
        self.assertEqual(conv.get_remote('key', 'default'), 'default')

        # no matching units
        related_units.side_effect = [['srv1/1'], []]
        relation_get.side_effect = AssertionError('relation_get should not be called')
        self.assertEqual(conv.get_remote('key', 'default'), 'default')

    @mock.patch.object(relations.hookenv, 'relation_get')
    @mock.patch.object(relations.hookenv, 'related_units')
    @mock.patch.object(relations.Conversation, 'relation_ids', ['rel:1',
                                                                'rel:2'])
    @mock.patch.object(relations.hookenv, 'hook_name',
                       lambda: 'rel-relation-departed')
    @mock.patch.object(relations.hookenv, 'relation_id', lambda: 'rel:2')
    @mock.patch.object(relations.hookenv, 'remote_unit', lambda: 'srv2/0')
    def test_get_remote_departed(self, related_units, relation_get):
        conv = relations.Conversation('rel',
                                      ['srv1/0', 'srv2/0'],
                                      'scope')

        related_units.side_effect = [['srv1/0'], ['srv2/0']]
        relation_get.side_effect = [None, 'value']
        self.assertEqual(conv.get_remote('key', 'default'), 'value')
        self.assertEqual(related_units.call_args_list, [mock.call('rel:1'),
                                                        mock.call('rel:2')])
        self.assertEqual(relation_get.call_args_list,
                         [mock.call('key', 'srv1/0', 'rel:1'),
                          mock.call('key', 'srv2/0', 'rel:2')])

    @mock.patch.object(relations.unitdata, 'kv')
    def test_set_local(self, kv):
        conv = relations.Conversation('rel', ['srv1/0', 'srv2/0', 'srv2/1'], 'scope')
        conv.set_local('foo', 'bar')
        conv.set_local(data={'foo': 'bar'})
        conv.set_local(foo='bar')
        conv.set_local('foo', 'bof', data={'foo': 'bar'}, foo='qux')
        self.assertEqual(kv().update.call_args_list, [
            mock.call({'foo': 'bar'}, prefix='reactive.conversations.rel.scope.local-data.'),
            mock.call({'foo': 'bar'}, prefix='reactive.conversations.rel.scope.local-data.'),
            mock.call({'foo': 'bar'}, prefix='reactive.conversations.rel.scope.local-data.'),
            mock.call({'foo': 'qux'}, prefix='reactive.conversations.rel.scope.local-data.'),
        ])
        kv().update.reset_mock()
        conv.set_local()
        assert not kv().update.called

    @mock.patch.object(relations.unitdata, 'kv')
    def test_get_local(self, kv):
        conv = relations.Conversation('rel', ['srv1/0', 'srv2/0', 'srv2/1'], 'scope')
        conv.get_local('foo', 'default')
        kv().get.assert_called_once_with('reactive.conversations.rel.scope.local-data.foo', 'default')


class TestMigrateConvs(unittest.TestCase):
    @mock.patch.object(relations, '_get_flag_value')
    @mock.patch.object(relations, 'set_flag')
    @mock.patch.object(relations, 'get_flags')
    @mock.patch.object(relations, 'hookenv')
    @mock.patch.object(relations.unitdata, 'kv')
    def test_migrate(self, kv, mhookenv, get_flags, set_flag, _get_flag_value):
        kv().getrange.side_effect = [
            {'reactive.conversations.rel:0.service': {
                'namespace': 'rel:0',
            }},
            {'reactive.conversations.rel.global': {
                'relation_name': 'rel',
                'scope': 'global',
                'units': ['service/0', 'service/1', 'service/3'],
            }},
            {'reactive.conversations.rel.service': {
                'relation_name': 'rel',
                'scope': 'service',
                'units': ['service/0', 'service/1', 'service/3'],
            }},
            {'reactive.conversations.rel.service/3': {
                'relation_name': 'rel',
                'scope': 'service/3',
                'units': ['service/3'],
            }},
        ]
        mhookenv.relation_ids.return_value = ['rel:1', 'rel:2']
        mhookenv.related_units.side_effect = [
            ['service/0', 'service/2'], ['service/3'],
            ['service/0', 'service/2'], ['service/3'],
        ]
        get_flags.side_effect = [['rel.joined', 'foo'], ['rel.joined', 'foo']]
        _get_flag_value.side_effect = [
            {'conversations': ['reactive.conversations.rel.service']},
            None,
            {'conversations': ['reactive.conversations.rel.service/3']},
            {'conversations': []},
        ]
        relations._migrate_conversations()
        assert not kv().set.called

        kv().set.reset_mock()
        relations._migrate_conversations()
        kv().set.assert_called_with('reactive.conversations.rel.global', {
            'namespace': 'rel',
            'scope': 'global',
            'units': ['service/0', 'service/1', 'service/3'],
        })
        assert not kv().unset.called
        assert not set_flag.called

        kv().set.reset_mock()
        relations._migrate_conversations()
        kv().set.assert_any_call('reactive.conversations.rel:1.service', {
            'namespace': 'rel:1',
            'scope': 'service',
            'units': ['service/0'],
        })
        kv().set.assert_called_with('reactive.conversations.rel:2.service', {
            'namespace': 'rel:2',
            'scope': 'service',
            'units': ['service/3'],
        })
        kv().unset.assert_called_with('reactive.conversations.rel.service')
        set_flag.assert_called_with('rel.joined', {'conversations': [
            'reactive.conversations.rel:1.service',
            'reactive.conversations.rel:2.service',
        ]})

        kv().set.reset_mock()
        relations._migrate_conversations()
        kv().set.assert_called_with('reactive.conversations.rel:2.service/3', {
            'namespace': 'rel:2',
            'scope': 'service/3',
            'units': ['service/3'],
        })
        kv().unset.assert_called_with('reactive.conversations.rel.service/3')
        set_flag.assert_called_with('rel.joined', {'conversations': [
            'reactive.conversations.rel:2.service/3',
        ]})


class TestRelationCall(unittest.TestCase):
    def setUp(self):
        self.r1 = mock.Mock(name='r1')
        from_name_p = mock.patch.object(relations, 'endpoint_from_name')
        self.from_name = from_name_p.start()
        self.addCleanup(from_name_p.stop)
        self.from_name.side_effect = lambda name: self.r1
        from_flag_p = mock.patch.object(relations, 'endpoint_from_flag')
        self.from_flag = from_flag_p.start()
        self.addCleanup(from_flag_p.stop)
        self.from_flag.side_effect = lambda name: self.r1

    def test_no_impl(self):
        self.r1 = None
        self.assertRaises(ValueError, relations.relation_call, 'method', relation_name='rel')
        self.assertRaises(ValueError, relations.relation_call, 'method', state='state')
        self.assertRaises(ValueError, relations.relation_call, 'method')

    def test_call_name(self):
        self.r1.method.return_value = 'result'
        result = relations.relation_call('method',
                                         'rel', None, None,
                                         'arg1', 'arg2')
        self.assertEqual(result, 'result')
        self.r1.method.assert_called_once_with('arg1', 'arg2')
        self.from_name.assert_called_once_with('rel')

    def test_call_flag(self):
        self.r1.method.return_value = 'result'
        result = relations.relation_call('method',
                                         None, 'flag', None,
                                         'arg1', 'arg2')
        self.assertEqual(result, 'result')
        self.r1.method.assert_called_once_with('arg1', 'arg2')
        self.from_flag.assert_called_once_with('flag')

    @mock.patch.object(relations, 'isinstance')
    def test_call_conversations(self, isinst):
        isinst.return_value = True
        self.r1.conversations.return_value = list(mock.Mock(scope=scope) for scope in ['s1', 's2'])
        result = relations.relation_call('conversations', 'rel')
        self.assertEqual(result, ['s1', 's2'])
        isinst.assert_called_once_with(self.r1, relations.RelationBase)



# Copyright 2018 Canonical Limited.
#
# This file is part of charms.reactive.
#
# charms.reactive is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charm-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent
import unittest
from unittest import mock

from charms import reactive


class TestTracer(unittest.TestCase):
    def test_nulltracer_api(self):
        self._test_api(reactive.trace.NullTracer())

    @mock.patch('charms.reactive.bus._short_action_id')
    @mock.patch('charms.reactive.bus.Handler.get_handlers')
    @mock.patch('charmhelpers.core.hookenv.log')
    def test_logtracer_api(self, log, gh, sid):
        sid.side_effect = lambda x, y: x
        h = reactive.bus.Handler
        gh.side_effect = [
            [h('handler_1')],
            [h('handler_2'), h('handler_3')],
        ]
        self._test_api(reactive.trace.LogTracer())
        # We are not wedded to this format. We should change it if we
        # can come up with something more generally readable. Multiline
        # strings are used to avoid hookenv.log call overhead (juju-log is
        # expensive), but that might be premature optimization.
        log.assert_has_calls([
            mock.call('tracer: starting handler dispatch, 0 flags set', 'DEBUG'),
            mock.call('tracer: hooks phase, 0 handlers queued', 'DEBUG'),
            mock.call(dedent('''\
                      tracer>
                      tracer: set flag flag_a
                      tracer: ++   queue handler handler_1
                      ''').strip(), 'DEBUG'),
            mock.call(dedent('''\
                      tracer>
                      tracer: cleared flag flag_b
                      tracer: ++   queue handler handler_2
                      tracer: ++   queue handler handler_3
                      tracer: -- dequeue handler handler_1
                      ''').strip(), 'DEBUG'),
        ])

    def _test_api(self, imp):
        imp.start_dispatch()
        imp.start_dispatch_phase('hooks', [])
        imp.start_dispatch_iteration(0, [])
        imp.set_flag('flag_a')
        imp.clear_flag('flag_b')



# Copyright 2014-2015 Canonical Limited.
#
# This file is part of charm-helpers.
#
# charm-helpers is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3 as
# published by the Free Software Foundation.
#
# charm-helpers is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with charm-helpers.  If not, see <http://www.gnu.org/licenses/>.

from charms.reactive import RelationBase


test_marker = 'hyphenated-relation'


class TestRelation(RelationBase):
    pass



from charmhelpers.core import hookenv

import charms.reactive


class NullTracer(object):
    """
    NullTracer is the default tracer and tracer base class, and does nothing

    This API will need to be adapted when charms.reactive internals
    change and cannot be considered stable.
    """
    def start_dispatch(self):
        """
        Start of handler dispatch
        """
        pass

    def start_dispatch_phase(self, phase, handlers):
        """
        Start of a particular dispatch phase. ie. hooks, restricted, or other.
        """
        pass

    def start_dispatch_iteration(self, iteration, handlers):
        """
        Start of an iteration of the dispatch phase.

        Only the 'other' phase (the main reactive loop) has multiple
        phases, and is the only phase that invokes this method.
        """

    def set_flag(self, flag):
        """
        A charms.reactive flag is being set.
        """
        pass

    def clear_flag(self, flag):
        """
        A charms.reactive flag is being cleared.
        """
        pass


class LogTracer(NullTracer):
    """
    LogTracer logs flags and handler activation to the Juju charm log

    Expect formatting and verbosity to change in future releases.
    """
    LEVEL = hookenv.DEBUG

    def __init__(self):
        self._active_handlers = set()
        self._msgs = []

    def start_dispatch(self):
        all_flags = charms.reactive.get_flags()
        self._emit("starting handler dispatch, {} flags set".format(len(all_flags)))
        for f in all_flags:
            self._emit("set flag {}".format(f))
        self._flush()

    def start_dispatch_phase(self, phase, handlers):
        self._active_handlers = set(handlers)
        if phase == 'other':
            self._emit("main dispatch loop, {} handlers queued".format(len(handlers)))
        else:
            self._emit("{} phase, {} handlers queued".format(phase, len(handlers)))
        for h in sorted(h.id() for h in handlers):
            self._emit("++   queue handler {}".format(h))
        self._flush()

    def start_dispatch_iteration(self, iteration, handlers):
        self._active_handlers = set(handlers)

    def set_flag(self, flag):
        self._flag("set flag {}".format(flag))

    def clear_flag(self, flag):
        self._flag("cleared flag {}".format(flag))

    def _emit(self, msg):
        self._msgs.append("tracer: {}".format(msg))

    def _flush(self):
        if self._msgs:
            if len(self._msgs) > 1:
                self._msgs.insert(0, "tracer>")
            hookenv.log("\n".join(self._msgs), self.LEVEL)
            self._msgs = []

    def _flag(self, msg):
        self._emit(msg)
        prev_handlers = self._active_handlers
        next_handlers = set(h for h in charms.reactive.bus.Handler.get_handlers() if h.test())

        for h in sorted(h.id() for h in (next_handlers - prev_handlers)):
            self._emit("++   queue handler {}".format(h))

        for h in sorted(h.id() for h in (prev_handlers - next_handlers)):
            self._emit("-- dequeue handler {}".format(h))

        self._flush()

        self._active_handlers = next_handlers


_tracer = None


def install_tracer(tracer):
    global _tracer
    _tracer = tracer
    # Disable tracing when we hit atexit, to avoid spam from layers
    # such as when the base layer tears down its automatic flags.
    hookenv.atexit(install_tracer, NullTracer())


def tracer():
    global _tracer
    return _tracer


# On import, the NullTracer is installed to avoid unit tests having
# to mock hookenv.log.
install_tracer(NullTracer())

# Install the LogTracer by default, after discovery, before running
# handlers. If this is too noisy for some, we can make it optional
# via layer.yaml. Using hookenv.atstart, because the tests already
# mock it.
hookenv.atstart(install_tracer, LogTracer())
